/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
namespace MpcSharp
{
    using System;
    using MPD;

    public sealed class Connection
    {
        static readonly Connection instance = new Connection();
        public static Connection Instance
        {
            get
            {
                return instance;
            }
        }

        /*
         * The constructor initializes all the status and stats info, making
         * sure we always have them non-null.
         */
        Connection()
        {
            statusInfo = new StatusInfo(new System.Collections.Hashtable());
            prevStatusInfo = new StatusInfo(new System.Collections.Hashtable());
            statsInfo = new StatsInfo(new System.Collections.Hashtable());
            prevStatsInfo = new StatsInfo(new System.Collections.Hashtable());
        }
        /*
         * The instance of the library communicating with mpd.
         */
        public  MPDClient  client;

        /*
         * All the status info returned from mpd.
         */
        public  StatusInfo statusInfo;
        public  StatusInfo prevStatusInfo;
        public  StatsInfo  statsInfo;
        public  StatsInfo  prevStatsInfo;
        public  MusicInfo  musicInfo;

        /*
         * This delegate is executed at each refresh.
         * Diffrent parts of the program can register delegates to be executed
         * at each refresh. As these are executed often (~2 times a second),
         * they should not contain any heavy code.
         * When a part of a program is removed at runtime (eg. a tab is
         * hidden), its update delegates must also be removed.
         */
        public delegate void UpdateDelegate();
        public UpdateDelegate UpdateDelegates;

        /*
         * Create the connection to mpd. Return true on success.
         */
        public bool Connect()
        {
            try
            {
                client = new MPDClient(Settings.Host, Settings.Port);
            }
            catch (MPDException)
            {
                return false;
            }
            catch (System.Net.Sockets.SocketException)
            {
                return false;
            }

            // Send the password if specified
            if (Settings.Password != "")
            {
                try
                {
                    client.Password(Settings.Password);
                }
                catch (MPDException)
                {
                    return false;
                }
            }
            return true;
        }

        /*
         * Process the status info returned from mpd.
         */
        void ResultCallback(object result)
        {
            if (result is StatusInfo)
            {
                prevStatusInfo = statusInfo;
                statusInfo = (StatusInfo) result;
            }
            else if (result is StatsInfo)
            {
                prevStatsInfo = statsInfo;
                statsInfo = (StatsInfo) result;
            }
            else if (result is MusicInfo)
                musicInfo = result as MusicInfo;
        }

        /*
         * Refreshes all the status and stats info from mpd. The last known
         * values are saved to be able to tell when interesting changes have
         * occured.
         */
        public bool Refresh()
        {
            try
            {
                // Update status info
                client.Status(ResultCallback);
                while(client.ReadResponse())
                {}

                // Update stats info
                client.Statistics(ResultCallback);
                while(client.ReadResponse())
                {}

                // Update info about currently playing song
                client.CurrentSong(ResultCallback);
                while(client.ReadResponse())
                {}
            }
            catch (MPDException e)
            {
                client.CleanCache();
                // Something has gone wrong, set client to null to force a new
                // connection to mpd.
                client = null;
                Console.WriteLine(e);
                return false;
            }
            catch (Exception e)
            {
                client = null;
                Console.WriteLine(e);
                return false;
            }
            return true;
        }
    }
}
