/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
using System;
using System.Collections;
using Gtk;

namespace MpcSharp
{
    /*
     * A tab displaying cover images.
     *
     * It constructs a list of images related to the playing song. The criteria
     * for a 'related' song is simply if it's located in the same directory.
     * One of the images is possibly selected as the 'front' image and one as
     * the 'back' image. The rest is regarded as 'other' image files related to
     * the playing song.
     */
    class Cover : ScrolledWindow, ITab, IUpdatable
    {
        /*
         * Only files with one of these file extensions are used.
         */
        enum ImageFormats
        {
            png,
            tiff,
            tif,
            jpg,
            jpeg,
            gif,
            pcx,
            ico,
            xpm,
            xbm
        }

        /*
         * The different zoom modes available.
         * NormalSize displays the image in 100% zoom.
         * BestFit resizes the image to fill the allocated space as good as
         * possible.
         */
        enum ZoomMode
        {
            NormalSize,
            BestFit
        }

        [Glade.Widget] MenuItem front;
        [Glade.Widget] MenuItem back;
        [Glade.Widget] MenuItem other;
        [Glade.Widget] MenuItem normal_size;
        [Glade.Widget] MenuItem best_fit;
        [Glade.Widget] Menu     cover_popup;
        Image image;



        // The dir in which the covers are
        string dir = "";

        /*
         * These are the filenames of the found covers.
         * The front and back covers are regarded as not found if they contain
         * only an empty string.
         */
        string frontFile;
        string backFile;
        ArrayList otherList = new ArrayList();

        Gdk.Pixbuf pixbuf;

        /*
         * For panning of the cover image. The position where the dragging of
         * the mouse must be saved. The same applies to the position of the
         * adjustment bars.
         */
        double oldXRoot;
        double oldYRoot;
        double oldHadj;
        double oldVadj;


        /*
         * Default to 'BestFit' mode.
         */
        ZoomMode mode = ZoomMode.BestFit;
        ZoomMode Mode
        {
            get
            {
                return mode;
            }
            set
            {
                mode = value;
                QueueResize();
            }
        }

        /*
         * When a new file should be shown, just store the filename in this Property.
         */
        string activeFile = "";
        string ActiveFile
        {
            get
            {
                return activeFile;
            }
            set
            {
                activeFile = value;
                LoadFile();
            }
        }

        public Cover() : base()
        {
            image = new Image();
            EventBox eventbox = new EventBox();
            eventbox.Visible = false;
            eventbox.Add(image);
            AddWithViewport(eventbox);
            SizeAllocated += new SizeAllocatedHandler(OnSizeAllocate);
            Mapped += new EventHandler(OnMapEvent);

            Glade.XML gxml_menu = Glade.XML.FromAssembly("cover.glade", "cover_popup", null);
            gxml_menu.Autoconnect(this);

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;

            // GConf notifications
            MpcSharp.Settings.MusicDirectoryChanged += MusicDirectoryNotifyEventHandler;
        }

        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        public TabType Type {
            get {
                return TabType.Cover;
            }
        }

        public string Title {
            get {
                return "Cover";
            }
        }

        void Update()
        {
            if (Connection.Instance.prevStatusInfo.Music.SongID !=
                    Connection.Instance.statusInfo.Music.SongID)
            {
                // New song -> Find the images
                dir = System.IO.Path.GetDirectoryName(Connection.Instance.musicInfo.Filename);
                FindImages();
            }
        }

        void MusicDirectoryNotifyEventHandler(object o, GConf.NotifyEventArgs args)
        {
            FindImages();
            QueueResize();
        }

        /*
         * Enumerate all the files in the current directory and extract the
         * images. Also choose one image as front image and one as back image
         * if found.
         * The priority, from high to low, when choosing which image to show is
         * front->back->other->none.
         */
        void FindImages()
        {
            // Zero previously found images
            ActiveFile = "";
            frontFile = "";
            backFile = "";
            otherList.Clear();

            if (MpcSharp.Settings.MusicDirectory == "" || dir == null)
                return;

            string absDir = System.IO.Path.Combine(MpcSharp.Settings.MusicDirectory, dir);
            string[] files;
            try
            {
                files = System.IO.Directory.GetFiles(absDir);
            }
            catch (System.IO.DirectoryNotFoundException)
            {
                // Coming here could indicate e.g. that a non existing
                // directory has been specified in the config or that we are
                // listening to a stream.
                return;
            }
            ArrayList images = new ArrayList();
            foreach (string file in files)
            {
                try
                {
                    ImageFormats.Parse(typeof(ImageFormats),
                            System.IO.Path.GetExtension(file).TrimStart('.'), true);
                    images.Add(file);
                }
                catch (ArgumentException)
                {
                    // Not a match - do nothing
                }
            }

            // Extract front and back images.
            foreach (string imageFile in images)
            {
                if (imageFile.ToLower().IndexOf("front") >= 0)
                {
                    if (frontFile == "")
                    {
                        frontFile = imageFile;
                        continue;
                    }
                }
                else if (imageFile.ToLower().IndexOf("back") >= 0)
                {
                    if (backFile == "")
                    {
                        backFile = imageFile;
                        continue;
                    }
                }
                otherList.Add(imageFile);
            }
            otherList.Sort();

            // Show 'best' image
            if (frontFile != "")
                ActiveFile = frontFile;
            else if (backFile != "")
                ActiveFile = backFile;
            else if (otherList.Count > 0)
                ActiveFile = otherList[0] as string;
        }

        /*
         * Load the actual image data from disk.
         */
        void LoadFile()
        {
            if (ActiveFile == "") // No cover file
                return;
            try
            {
                pixbuf = new Gdk.Pixbuf(ActiveFile);
                image.Pixbuf = pixbuf;
            }
            catch (GLib.GException)
            {
                Console.WriteLine("Cover: File not found: '{0}'", ActiveFile);
            }
            QueueResize();
        }

        /*
         * Main gui callbacks
         */

        /*
         *  Ensure that the image is resized when the window is resized.
         */
        protected void OnSizeAllocate(object o, System.EventArgs args)
        {
            // Abort if the covertab is not visible
            if (Parent is Notebook)
            {
                if ((Parent as Notebook).CurrentPageWidget != this)
                {
                    return;
                }
            }

            // Show a small default transparent image when no image is
            // available
            if (ActiveFile == "")
            {
                image.Pixbuf = new Gdk.Pixbuf(Gdk.Colorspace.Rgb, true, 8, 1, 1);
                image.Pixbuf.Fill(0);
                return;
            }

            if (Mode == ZoomMode.BestFit)
            {
                // Rescale the image, filling the allocated area while
                // preserving the aspect ratio.
                double width  = Hadjustment.PageSize;
                double height = Vadjustment. PageSize;
                double viewAR = width / height;
                double pixbufAR = pixbuf.Width / (double) pixbuf.Height;
                int newWidth, newHeight;
                if (viewAR < pixbufAR)
                {
                    newWidth = Convert.ToInt32(width);
                    newHeight = Convert.ToInt32(width / pixbufAR);
                }
                else
                {
                    newWidth = Convert.ToInt32(height * pixbufAR);
                    newHeight = Convert.ToInt32(height);
                }
                image.Pixbuf = 
                    pixbuf.ScaleSimple(newWidth, newHeight, Gdk.InterpType.Bilinear);
            }
            else // Mode == ZoomMode.NormalSize
            {
                image.Pixbuf = pixbuf;
            }
        }

        /*
         * Mouse button 3 popups the menu.
         * The menu containing the 'other' image files is constructed on the
         * fly.
         *
         * Mouse button 1 might indicate that a pan has begin and we must save
         * the initial coordinates.
         */
        [GLib.ConnectBefore]
        protected override bool OnButtonPressEvent(Gdk.EventButton evnt)
        {
            if (evnt.Button == 3)
            {
                front.Sensitive = frontFile != "";
                back.Sensitive = backFile != "";
                other.Sensitive = otherList.Count > 0;
                normal_size.Sensitive = ActiveFile != "";
                best_fit.Sensitive = ActiveFile != "";

                // Construct menu for 'other' covers
                if (otherList.Count > 0)
                {
                    Menu menu = new Menu();
                    other.Submenu = menu;
                    foreach (string file in otherList)
                    {
                        string label = System.IO.Path.GetFileName(file);
                        MenuItem item = new PayloadMenuItem(label, file);
                        menu.Append(item);
                        item.Activated += new EventHandler(OnMenuOther);
                    }
                    menu.ShowAll();
                }
                cover_popup.Popup();
                return true;
            }
            else if (evnt.Button == 1)
            {
                oldXRoot = evnt.XRoot;
                oldYRoot = evnt.YRoot;
                oldHadj  = Hadjustment.Value;
                oldVadj  = Vadjustment.Value;
            }
            return true;
        }

        /*
         * Pan the displayed cover image with mouse button 1.
         */
        protected override bool OnMotionNotifyEvent(Gdk.EventMotion evnt)
        {
            // We must drag with mouse button 1...
            if ((evnt.State & Gdk.ModifierType.Button1Mask) !=
                    Gdk.ModifierType.Button1Mask)
                return false;
            // ...and be in 'NormalSize' mode
            if (Mode != ZoomMode.NormalSize)
                return false;

            // Calculate new Adjustment values based on where we began the dragging
            // motion.
            Adjustment hadj = Hadjustment;
            Adjustment vadj = Vadjustment;
            double dx = evnt.XRoot - oldXRoot;
            double dy = evnt.YRoot - oldYRoot;
            double newHadj = System.Math.Min(oldHadj - dx, hadj.Upper - hadj.PageSize);
            double newVadj = System.Math.Min(oldVadj - dy, vadj.Upper - vadj.PageSize);
            hadj.Value = newHadj;
            vadj.Value = newVadj;
            return false;
        }

        void OnMapEvent(object o, EventArgs args)
        {
            QueueResize();
        }

        /*
         * Menu callbacks
         */
        private void OnMenuFront(object o, EventArgs args)
        {
            ActiveFile = frontFile;
        }

        private void OnMenuBack(object o, EventArgs args)
        {
            ActiveFile = backFile;
        }

        private void OnMenuOther(object o, EventArgs args)
        {
            ActiveFile = (o as PayloadMenuItem).Payload as string;
        }

        private void OnMenuNormalSize(object o, EventArgs args)
        {
            Mode = ZoomMode.NormalSize;
        }

        private void OnMenuBestFit(object o, EventArgs args)
        {
            Mode = ZoomMode.BestFit;
        }

    }
    /*
     * Extension of Gtk.MenuItem that can store an 'object' payload.
     */
    class PayloadMenuItem : Gtk.MenuItem
    {
        object payload;
        public object Payload
        {
            get
            {
                return payload;
            }
        }
        public PayloadMenuItem(string label, object payload) : base(label)
        {
            this.payload = payload;
        }
    }
}
