/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using Gtk;
using MPD;

namespace MpcSharp
{
    /*
     * A tab displaying search results.
     *
     * A search is defined by a search string and a metadata field.
     * All songs where the selected metadata field contains the search string
     * is displayed - no exact match is required.
     * Searches are case insensitive.
     */
    class Find : VBox, ITab
    {
        Glade.XML gxml;
        ListStore tree;
        ComboBox combobox;

        [Glade.Widget] TreeView view;
        [Glade.Widget] Entry entry;
        [Glade.Widget] HBox hbox;

        [Glade.Widget] MenuItem menu_play;
        [Glade.Widget] MenuItem menu_add;
        [Glade.Widget] Menu find_popup;

        public Find() : base()
        {
            gxml = Glade.XML.FromAssembly("find.glade", "hbox", null); 
            gxml.Autoconnect(this);
            PackStart(gxml.GetWidget("hbox"), false, false, 0);

            gxml = Glade.XML.FromAssembly("find.glade", "scrolled_win", null); 
            gxml.Autoconnect(this);
            PackStart(gxml.GetWidget("scrolled_win"));


            Glade.XML gxml_menu = Glade.XML.FromAssembly("find.glade", "find_popup", null); 
            gxml_menu.Autoconnect(this);

            tree = new ListStore(typeof(string));
            view.Model = tree;
            view.AppendColumn("Results", new CellRendererText(), "text", 0);
            view.Columns[0].Sizing = TreeViewColumnSizing.Autosize;
            view.Selection.Mode = SelectionMode.Multiple;

            // Fill combo box with metadata types
            combobox = ComboBox.NewText();
            foreach (string item in SearchType.GetNames(typeof(SearchType)))
            {
                combobox.AppendText(item);
            }
            combobox.Active = 0;
            hbox.PackStart(combobox, false, false, 0);
            combobox.ShowAll();
            combobox.Changed += new EventHandler(OnChanged);
        }

        public TabType Type {
            get {
                return TabType.Find;
            }
        }

        public string Title {
            get {
                return "Find";
            }
        }


        /*
         * Process the search results returned from mpd, ie. show them in the
         * view.
         */
        void SearchResultsCb(object result)
        {
            // Just add all matching songs to the ListStore
            foreach (MusicInfo info in (result as MusicCollection).Songs)
            {
                tree.AppendValues(info.Filename);
            }
        }

        /*
         * Take all the selected rows and add them to the playlist.
         */
        void EnqueueTreeSelection()
        {
            TreeIter iter;
            // Return if nothing is selected
            if (!tree.GetIterFirst(out iter))
                return;

            TreeModel model;
            TreePath[] paths = view.Selection.GetSelectedRows(out model);

            try
            {
                foreach (TreePath path in paths)
                {
                    model.GetIter(out iter, path);
                    string val = (string) model.GetValue(iter, 0);
                    if (Connection.Instance.client != null)
                    {
                        Connection.Instance.client.Add(val);
                    }
                }
                while (Connection.Instance.client.ReadResponse()) {}
            }
            catch (MPD.MPDException)
            {}
        }

        void ShowMenu()
        {
            bool selCount = view.Selection.CountSelectedRows() > 0;
            menu_play.Sensitive = selCount;
            menu_add.Sensitive = selCount;

            find_popup.Popup();
        }

        /*
         * Main gui callbacks
         */
        void OnChanged(object o, EventArgs args)
        {
            if (MpcSharp.Settings.IncrementalSearch)
                entry.Activate();
        }

        /*
         * The search entry box has been activated - make a new search.
         */
        void OnActivate(object o, EventArgs args)
        {
            // Clear old results
            tree.Clear();

            string what = entry.Text;

            // Require at least two characters
            if (what.Length < 2)
                return;
            try
            {
                TreeIter iter;
                combobox.GetActiveIter(out iter);
                string text = (string) combobox.Model.GetValue(iter, 0);
                
                SearchType type = (SearchType) SearchType.Parse(
                        typeof(SearchType), text, true);
                // Send command to mpd
                if (what != String.Empty)
                {
                    Connection.Instance.client.Search(type, what, false, SearchResultsCb);
                    while (Connection.Instance.client.ReadResponse())
                    {}
                }
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
                Connection.Instance.client.CleanCache();
            }
            catch (Exception e)
            {
                Console.WriteLine(e);
            }
        }

        void OnRowActivated(object o, RowActivatedArgs args){
            EnqueueTreeSelection();
        }

        /*
         * If we only have one selected item, a right click will deselect it
         * and select the item under the cursor.
         * If more than one item is selected, do not deselect them at a right
         * click.
         */
        [GLib.ConnectBefore]
        void OnButtonPress(object o, ButtonPressEventArgs args){
            if (args.Event.Button == 3)
                args.RetVal = (o as TreeView).Selection.CountSelectedRows() > 1;
        }

        void OnButtonRelease(object o, ButtonReleaseEventArgs args){
            if (args.Event.Button == 3)
                ShowMenu();
            args.RetVal = true;
        }
        
        void OnPopupMenu(object o, PopupMenuArgs args){
            ShowMenu();
        }

        /*
         * Menu callbacks
         */
        void OnMenuPlay(object o, EventArgs args)
        {
            // To 'Play' items is equivalent to clear and then 'Add' them.
            try
            {
                if (Connection.Instance.client != null)
                {
                    Connection.Instance.client.Clear();
                    Connection.Instance.client.ReadResponse();
                }
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            menu_add.Activate();
        }

        void OnMenuAdd(object o, EventArgs args)
        {
            EnqueueTreeSelection();
            try
            {
                if (Connection.Instance.client != null)
                {
                    Connection.Instance.client.Play();
                    Connection.Instance.client.ReadResponse();
                }
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
        }
    }
}
