/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using Gtk;
using MPD;

/*
 * The two rows with control widgets are all derived from Gtk.Widget.
 * This file contains all these available widgets.
 * They are all completely independent which makes it easy to add or remove
 * them independently at runtime.
 */

namespace MpcSharp
{

    class Factory
    {
        /*
         * Add a GuiWidget to the specified container.
         */
        public static void AddGuiWidget(string name, Box box)
        {
            if (name == "")
                return;
            Widget w = null;
            GuiWidget widget = (GuiWidget) Enum.Parse(typeof(GuiWidget), name);
            switch (widget)
            {
                case GuiWidget.Prev:
                    w = new PrevButton();
                    break;
                case GuiWidget.Play:
                    w = new PlayButton();
                    break;
                case GuiWidget.Stop:
                    w = new StopButton();
                    break;
                case GuiWidget.Next:
                    w = new NextButton();
                    break;
                case GuiWidget.Random:
                    w = new RandomButton();
                    break;
                case GuiWidget.Repeat:
                    w = new RepeatButton();
                    break;
                case GuiWidget.Time:
                    w = new TimeLabel();
                    break;
            }
            if (w != null)
            {
                box.PackStart(w, false, false, 0);
                return;
            }

            switch (widget)
            {
                case GuiWidget.Progress:
                    w = new ProgressScale();
                    break;
                case GuiWidget.Volume:
                    w = new VolumeScale();
                    break;
            }
            if (w != null)
                box.PackStart(w, true, true, 0);
        }


    }


    /*
     * The diffrent time formats available.
     */
    public enum TimeFormat
    {
        Elapsed,
        Remaining
    }

    /*
     * A list of all the available widgets. These names (converted to strings)
     * are saved in the GConf settings.
     */
    public enum GuiWidget
    {
        Prev,
        Play,
        Stop,
        Next,
        Volume,
        Random,
        Repeat,
        Time,
        Progress
    }

    /*
     * A convenience class defining some common properties for a widget with
     * one button showing an icon.
     */
    abstract class MpcSharpButton : Button
    {
        private Image image;
        protected string Resource
        {
            set
            {
                Gdk.Pixbuf pixbuf = new Gdk.Pixbuf(null, value);
                image.Pixbuf = pixbuf;
                image.ShowAll();
            }
        }

        public MpcSharpButton() : base()
        {
            Relief = ReliefStyle.None;
            image = new Image();
            Add(image);
        }
    }

    /*
     * A convenience class defining some common properties for a widget with
     * a toggle button showing two diffrent icons depending on its
     * active/inactive state.
     */
    abstract class MpcSharpToggleButton : ToggleButton
    {
        protected Gdk.Pixbuf pixbufActive;
        protected Gdk.Pixbuf pixbufInactive;
        protected Image image;

        protected string ActiveResource
        {
            set
            {
                pixbufActive = new Gdk.Pixbuf(null, value);
            }
        }

        protected string InactiveResource
        {
            set
            {
                pixbufInactive = new Gdk.Pixbuf(null, value);
            }
        }

        public MpcSharpToggleButton() : base()
        {
            Relief = ReliefStyle.None;
            image = new Image();
            Add(image);
        }
    }

    /*
     * When clicked, send the 'prev' command to mpd.
     */
    class PrevButton : MpcSharpButton
    {
        public PrevButton() : base()
        {
            Resource = "prev.png";
        }

        override protected void OnClicked()
        {
                try
                {
                    Connection.Instance.client.Prev();
                    Connection.Instance.client.ReadResponse();
                }
                catch (MPD.MPDException e)
                {
                    Console.WriteLine(e);
                }
                catch (NullReferenceException)
                {}
        }
    }

    /*
     * When clicked, toggle mpd between 'playing' and 'stopped' state.
     */
    class PlayButton : MpcSharpButton
    {
        public PlayButton() : base()
        {
            Resource = "play.png";
        }

        override protected void OnClicked()
        {
            StateStatus state = Connection.Instance.statusInfo.State;
            try
            {
                if (state == StateStatus.Stop)
                    Connection.Instance.client.Play();
                else
                    Connection.Instance.client.Paused(state == StateStatus.Play ? true : false);
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }


    /*
     * When clicked, stop playing.
     */
    class StopButton : MpcSharpButton
    {
        public StopButton() : base()
        {
            Resource = "stop.png";
        }

        override protected void OnClicked()
        {
            try
            {
                Connection.Instance.client.Stop();
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    /*
     * When clicked, send the 'next' command to mpd.
     */
    class NextButton : MpcSharpButton
    {
        public NextButton() : base()
        {
            Resource = "next.png";
        }

        override protected void OnClicked()
        {
            try
            {
                Connection.Instance.client.Next();
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    /*
     * For controlling the volume level.
     */
    class VolumeScale : HScale, IUpdatable
    {
        public VolumeScale() : base(0, 100, 2)
        {
            ValuePos = PositionType.Left;

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;
        }
        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        void Update()
        {
            if (Value != Connection.Instance.statusInfo.Volume)
            {
                Value = Connection.Instance.statusInfo.Volume;
            }
        }

        override protected void OnValueChanged()
        {
            try
            {
                Connection.Instance.client.Volume((int) Value);
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        override protected string OnFormatValue(double val)
        {
            return "Volume:";
        }


    }

    /*
     * When clicked, toggle the mpd random state.
     */
    class RandomButton : MpcSharpToggleButton, IUpdatable
    {

        public RandomButton() : base()
        {
            ActiveResource = "random_on.png";
            InactiveResource = "random_off.png";
            image.Pixbuf = pixbufActive;
            Toggled += new EventHandler(OnToggled);

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;
        }

        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        void Update()
        {
            Toggled -= new EventHandler(OnToggled);
            Active = Connection.Instance.statusInfo.Random;
            Toggled += new EventHandler(OnToggled);
            if (Active)
                image.Pixbuf = pixbufActive;
            else
                image.Pixbuf = pixbufInactive;
        }

        void OnToggled(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Random(!Connection.Instance.statusInfo.Random);
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    /*
     * When clicked, toggle the mpd repeat state.
     */
    class RepeatButton : MpcSharpToggleButton, IUpdatable
    {

        public RepeatButton() : base()
        {
            ActiveResource = "repeat.png";
            Toggled += new EventHandler(OnToggled);
            image.Pixbuf = pixbufActive;

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;
        }

        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        void Update()
        {
            Toggled -= new EventHandler(OnToggled);
            Active = Connection.Instance.statusInfo.Repeat;
            Toggled += new EventHandler(OnToggled);
        }

        void OnToggled(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Repeat(!Connection.Instance.statusInfo.Repeat);
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }

    /*
     * Show the current or the elapsed time and the total time of the currently
     * playing song.
     * When clicked, switch between the diffrent TimeFormats.
     */
    class TimeLabel : EventBox, IUpdatable
    {
        Label label;

        public TimeLabel() : base()
        {
            label = new Label();
            Add(label);

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;
        }

        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        void Update()
        {
            int elapsed = Connection.Instance.statusInfo.Music.ElapsedTime;
            int total   = Connection.Instance.statusInfo.Music.TotalTime;
            if (elapsed < 0)
                label.Text = "00:00 / 00:00";
            else if (MpcSharp.Settings.TimeFormat == TimeFormat.Elapsed)
            {
                label.Text = Format.TimeInSeconds(elapsed) + " / " +
                    Format.TimeInSeconds(total);
            }
            else if (MpcSharp.Settings.TimeFormat == TimeFormat.Remaining)
            {
                int remaining = System.Math.Abs(total - elapsed);
                label.Text = "-" + Format.TimeInSeconds(remaining) + " / " +
                    Format.TimeInSeconds(total);
            }
        }

        override protected bool OnButtonPressEvent(Gdk.EventButton evnt)
        {
            if (MpcSharp.Settings.TimeFormat == TimeFormat.Elapsed)
                MpcSharp.Settings.TimeFormat = TimeFormat.Remaining;
            else
                MpcSharp.Settings.TimeFormat = TimeFormat.Elapsed;
            return true;
        }
    }

    /*
     * A progress scale showing the current position in the playing song. Drag
     * the scale to seek.
     */
    class ProgressScale : HScale, IUpdatable
    {
        public ProgressScale() : base(0, 1, 2)
        {
            DrawValue = false;

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;
        }

        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        void Update()
        {
            int elapsed = Connection.Instance.statusInfo.Music.ElapsedTime;
            int total   = Connection.Instance.statusInfo.Music.TotalTime;
            if (total > 0)
            {
                SetRange(0, total);
                Value = elapsed;
            }
            else
                Value = 0;
        }

        override protected void OnAdjustBounds(double new_value)
        {
            UpdatePolicy = Gtk.UpdateType.Discontinuous;
            try
            {
                Connection.Instance.client.SeekTo(Connection.Instance.statusInfo.Music.SongID,
                        Convert.ToInt32(new_value));
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }
}
