/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using Gtk;
using MPD;


namespace MpcSharp
{
    /*
     * These are the diffrent types a row in the library can be. It's used to
     * know what to do when activating rows.
     */
    enum EntityType
    {
        File,
        Dir,
        Parent,
        Playlist
    }

    /*
     * A tab showing all songs in the music library.
     *
     * The music is browsable in two diffrent main ways:
     *  - By filename:
     *      In this mode, the whole directory tree specified as the
     *      music_directory in the mpd configuration file can be navigated.
     *  - By metadata:
     *      In this mode, one 'directory' for each unique metadata field of the
     *      chosen typeis created. Its content can be browsed as in filename
     *      mode.
     *      Here we never have more than one level of directories.
     *
     * The library is also used when updating the mpd database file.
     */
    class Library : VBox, ITab, IUpdatable
    {
        System.Collections.Stack PositionStack = new System.Collections.Stack();
        Glade.XML gxml_menu;
        ListStore tree;
        System.Collections.Stack DirStack = new System.Collections.Stack();
        string DirStackString = "";
        [Glade.Widget] MenuItem menu_browse_by;
        [Glade.Widget] MenuItem menu_update_this_dir;
        [Glade.Widget] MenuItem menu_add;
        [Glade.Widget] MenuItem menu_play;

        Label cwd_label = new Label();
        TreeView view = new TreeView();
        SearchType browseType;


        public Library() : base()
        {
            ScrolledWindow scrolled = new ScrolledWindow();

            // Pack widgets
            PackStart(cwd_label, false, false, 3);
            PackStart(scrolled);
            scrolled.Add(view);

            cwd_label.Xalign = 0;
            cwd_label.Xpad = 4;
            view.HeadersVisible = false;

            view.RowActivated += new RowActivatedHandler(OnRowActivated);
            view.ButtonPressEvent += new ButtonPressEventHandler(OnButtonPress);
            view.ButtonReleaseEvent += new ButtonReleaseEventHandler(OnButtonRelease);
            view.PopupMenu += new PopupMenuHandler(OnPopupMenu);

            gxml_menu = Glade.XML.FromAssembly("library.glade", "lib_popup", null);
            gxml_menu.Autoconnect(this);

            // Construct the 'Browse by' menu
            Menu menu = new Menu();
            menu_browse_by.Submenu = menu;

            foreach (SearchType type in Enum.GetValues(typeof(SearchType)))
            {
                // Only show items from 'SearchType' that are available in
                // 'StoreCols'. The item 'Any' alone does not fulfill this
                // requirement.
                MenuItem item = new PayloadMenuItem(type.ToString(), type);
                if (type != SearchType.Any)
                {
                    menu.Append(item);
                    item.Activated += new EventHandler(OnMenuBrowseBy);

                }
                menu.ShowAll();
            }

            tree = new ListStore(typeof(string), typeof(int));

            view.Model = tree;
            CellRenderer cellPixbuf = new CellRendererPixbuf();

            view.AppendColumn("Type", cellPixbuf);
            view.Columns[0].SetCellDataFunc(cellPixbuf, 
                    new TreeCellDataFunc(cellDataFunc));
            view.AppendColumn("Library", new CellRendererText(), "text", 0);
            view.Columns[1].Sizing = TreeViewColumnSizing.Autosize;
            view.Selection.Mode = SelectionMode.Multiple;

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;
        }

        /*
         * Displays different icons in the view depending on what type of entry
         * it is; a file, a directory or the parent directory entry.
         */
        void cellDataFunc (TreeViewColumn tvc, CellRenderer cell, TreeModel model, TreeIter iter)
        {
            int type = (int) model.GetValue(iter, 1);
            if (type == (int) EntityType.File)
                (cell as CellRendererPixbuf).StockId = Gtk.Stock.New;
            else if (type == (int) EntityType.Dir)
                (cell as CellRendererPixbuf).StockId = Gtk.Stock.Open;
            else if (type == (int) EntityType.Parent)
                (cell as CellRendererPixbuf).StockId = "";
        }


        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        public TabType Type {
            get {
                return TabType.Library;
            }
        }

        public string Title {
            get {
                return "Library";
            }
        }

        // This function is called once for every poll to mpd
        void Update()
        {
            if (Connection.Instance.statsInfo.DBUpdateDate > Connection.Instance.prevStatsInfo.DBUpdateDate)
                Reload();
        }

        /*
         * Convert the current directory stack to a string path.
         * The stack 'foo->bar' is converted to the string 'foo/bar'.
         * We can't store the path directly since we also browse a metadata
         * hierarcy where slashes are allowed.
         */
        string GetDirStackString()
        {
            if (DirStack.Count <= 0)
                return "";
            string dir = "";
            object [] list = DirStack.ToArray();
            Array.Reverse(list);
            foreach(object o in list)
            {
                dir += (string) o + "/";
            }
            return dir.Substring(0, dir.Length-1);
        }

        /*
         * (Re)read all info required to display the view of the library from
         * mpd.
         * It's called
         *  - automatically when a new version of the database is detected.
         *  - if a new directory is entered.
         *  - when choosing a new metadata field to browse by.
         */
        void Reload()
        {
            tree.Clear();
            cwd_label.Text = DirStackString;

            // Append parent dir entry if not at top already
            if (DirStackString != "")
            {
                tree.AppendValues("..", (int) EntityType.Parent);
            }

            // The two cases are 'browse by filename' and 'browse by alla others'
            try
            {
                if (browseType == SearchType.Filename)
                {
                    // Reload directories
                    Connection.Instance.client.ListDirectories(DirStackString, ListDirectoriesCallback);
                    while (Connection.Instance.client.ReadResponse())
                    {}

                    // Reload files
                    Connection.Instance.client.ListFiles(DirStackString, ListFilesCallback);
                    while (Connection.Instance.client.ReadResponse())
                    {}
                }
                else // Browse by some metadata field
                {
                    if (DirStackString == "")
                    {
                        Connection.Instance.client.ListByMetadata(browseType, ListByMetadataCallback);
                        while (Connection.Instance.client.ReadResponse())
                        {}
                    }
                    else
                    {
                        Connection.Instance.client.ListFilenamesByMetadata(
                                browseType, DirStackString, ListFilenamesByMetadataCallback);
                        while (Connection.Instance.client.ReadResponse())
                        {}
                    }
                }
            }
            catch (MPD.MPDException e)
            {
                Connection.Instance.client.CleanCache();
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}

        }

        void ListDirectoriesCallback(object result)
        {
            foreach (string dir in (string[]) result)
            {
                string[] temp = dir.Split('/');
                tree.AppendValues(temp[temp.Length - 1], (int) EntityType.Dir);
            }
        }

        void ListFilesCallback(object result)
        {
            foreach (string file in (string[]) result)
            {
                tree.AppendValues(System.IO.Path.GetFileName(file), (int) EntityType.File);
            }
        }

        void ListByMetadataCallback(object result)
        {
            foreach (string entity in (string[]) result)
            {
                tree.AppendValues(entity, (int) EntityType.Dir);
            }
        }

        void ListFilenamesByMetadataCallback(object result)
        {
            foreach (string entity in (string[]) result)
            {
                tree.AppendValues(entity, (int) EntityType.File);
            }
        }

        void EnqueueFilenamesByMetadataCallback(object result)
        {
            foreach (string entity in (string[]) result)
            {
                try
                {
                    Connection.Instance.client.Add(entity);
                }
                catch (MPD.MPDException e)
                {
                    Connection.Instance.client.CleanCache();
                    Console.WriteLine(e);
                }
                catch (NullReferenceException)
                {}
            }
        }

        /*
         * This is executed whenever the user double clicks a row.
         * If a directory is activated, the contents of that directory replaces
         * the contents of the view.
         * If a file is activated, that file is appended to the playlist.
         */
        void ActivateRow(TreePath treePath )
        {
            TreeIter iter = new TreeIter();
            tree.GetIter(out iter, treePath);

            string selected = (string) tree.GetValue(iter, 0);

            int type = (int) tree.GetValue(iter, 1);
            //char delim = System.IO.Path.DirectorySeparatorChar;

            // Dir activated
            if (type == (int) EntityType.Dir)
            {
                PositionStack.Push(treePath.ToString());
                DirStack.Push(selected);
                DirStackString = GetDirStackString();
                Reload();
                view.ScrollToCell(new TreePath("0"), null, true, 0.5f, 0.0f);
            }
            // Parent dir activated
            else if (type == (int) EntityType.Parent)
            {
                try
                {
                    DirStack.Pop();
                    DirStackString = GetDirStackString();
                }
                catch (System.ArgumentOutOfRangeException)
                {
                    DirStack.Clear();
                    DirStackString = GetDirStackString();
                }
                Reload();
                TreePath prevPath = new TreePath(PositionStack.Pop().ToString());
                view.SetCursor(prevPath, view.Columns[0], false);
                view.ScrollToCell(prevPath, null, true, 0.5f, 0.0f);
            }
            // File activated
            else if (type == (int) EntityType.File)
            {
                try
                {
                    if (browseType == SearchType.Filename)
                        Connection.Instance.client.Add(System.IO.Path.Combine(DirStackString, selected));
                    else
                        Connection.Instance.client.Add(selected);
                    Connection.Instance.client.ReadResponse();
                }
                catch (MPD.MPDException)
                {}
                catch (NullReferenceException)
                {}
            }
        }

        /*
         * Take all the selected rows and add them to the playlist. If a row is
         * a file, it's appended directly.
         * If it's a directory, the directory is recursively added.
         * When adding directories in metadata browse mode, each directory's
         * contents must first be queried from mpd.
         */
        void EnqueueSelection()
        {
            TreeIter iter;
            // Return if nothing is selected
            if (!tree.GetIterFirst(out iter))
                return;

            TreeModel model;
            TreePath[] paths = view.Selection.GetSelectedRows(out model);

            foreach (TreePath path in paths)
            {
                model.GetIter(out iter, path);
                EntityType type = (EntityType) model.GetValue(iter, 1);
                try
                {
                    string val = (string) model.GetValue(iter, 0);
                    if (browseType == SearchType.Filename)
                    {
                        if (type == EntityType.File || type == EntityType.Dir)
                        {
                            Connection.Instance.client.Add(System.IO.Path.Combine(DirStackString, val));
                        }
                    }
                    else // Metadata browse mode
                    {
                        if (type == EntityType.File)
                        {
                            Connection.Instance.client.Add(val);
                        }
                        else if (type == EntityType.Dir)
                        {
                            Connection.Instance.client.ListFilenamesByMetadata(
                                    browseType, val, EnqueueFilenamesByMetadataCallback);
                        }
                    }
                }
                catch (MPD.MPDException e)
                {
                    Console.WriteLine(e);
                }
                catch (NullReferenceException)
                {}
            }
            while (Connection.Instance.client.ReadResponse()) {}
        }

        /*
         * Show the popup menu. Also make sure only valid choises are available
         * for selection.
         */
        void ShowMenu()
        {
            bool selCount = view.Selection.CountSelectedRows() > 0;
            menu_play.Sensitive = selCount;
            menu_add.Sensitive = selCount;
            menu_update_this_dir.Sensitive = browseType == SearchType.Filename;

            (gxml_menu.GetWidget("lib_popup") as Menu).Popup();
        }

        /*
         * Main gui callbacks
         */
        void OnRowActivated(object o, RowActivatedArgs args)
        {
            ActivateRow(args.Path);
        }

        [GLib.ConnectBefore]
        void OnButtonPress(object o, ButtonPressEventArgs args)
        {
            if (args.Event.Button == 3)
            {
                // If multiple items selected, don't deselect them upon menu popup
                args.RetVal =  ((TreeView) o).Selection.CountSelectedRows() > 1;
            }
        }

        void OnButtonRelease(object o, ButtonReleaseEventArgs args)
        {
            if (args.Event.Button == 3)
            {
                ShowMenu();
            }
            else if (args.Event.Button == 2)
            {
                // Goto parent dir if not already in the root node
                TreeIter iter;
                if (tree.GetIterFirst(out iter))
                    if ((EntityType) tree.GetValue(iter, 1) == EntityType.Parent)
                        ActivateRow(new TreePath("0"));
            }
            args.RetVal = true;
        }

        void OnPopupMenu(object o, PopupMenuArgs args)
        {
            ShowMenu();
        }

        /*
         * Menu callbacks
         */
        void OnMenuPlay(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Clear();
                EnqueueSelection();
                Connection.Instance.client.Play();
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        void OnMenuAdd(object o, EventArgs args)
        {
            EnqueueSelection();
            try
            {
                Connection.Instance.client.Play();
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        void OnMenuBrowseBy(object o, EventArgs args)
        {
             browseType = (SearchType) (o as PayloadMenuItem).Payload;
             DirStack.Clear();
             DirStackString = GetDirStackString();
             PositionStack.Clear();
             Reload();
        }

        void OnMenuUpdateAll(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Update();
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        void OnMenuUpdateThisDir(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Update(DirStackString);
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

    }
}
