/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using Gtk;
using MPD;
using System.Collections;
using GConf;

namespace MpcSharp
{
    /*
     * All the columns that can be displayed.
     */
    public enum PlaylistColumn
    {
        Filename,
        Artist,
        Album,
        Title,
        Genre,
        Date,
        Track,
        Composer,
        Performer,
        Comment,
        Time,
        Custom
    }

    /*
     * A tab showing the current playlist.
     *
     * The data beneath the playlist view is stored in a ListStore. A TreeView
     * is used to show this data. The columns in the TreeView can be configured
     * dynamically at runtime. The available columns are
     *  - filename
     *  - any of the metadata fields (eg. Album or Genre)
     *  - a custom column with more elaborate rendering. This column have the
     *    same configurabili as the status string. Since you have to parse all
     *    the metadata for every item in the playlist and render the pango
     *    string, this can become slow for large playlists.
     */
    class Playlist : VBox, ITab, IUpdatable
    {
        Glade.XML gxml;
        Glade.XML gxml_menu;
        ListStore tree;
        CellRendererText cell;
        [Glade.Widget] TreeView view;
        TreeModelFilter filter;
        Filter flt;

        /*
         * These are the data that we store in the ListStore.
         */
        enum StoreCols
        {
            Filename = 0,
            Artist,
            Album,
            Title,
            Genre,
            Date,
            Track,
            Composer,
            Performer,
            Comment,
            Time,
            ID,
            IsPlaying
        }


        public Playlist() : base()
        {
            cell  = new CellRendererText();

            gxml = Glade.XML.FromAssembly("playlist.glade", "scrolled_win", null);
            gxml.Autoconnect(this);
            PackStart(gxml.GetWidget("scrolled_win"));

            gxml_menu = Glade.XML.FromAssembly("playlist.glade", "pl_popup", null);
            gxml_menu.Autoconnect(this);

            tree = new ListStore(
                    typeof(string), // Filename
                    typeof(string), // Artist
                    typeof(string), // Album
                    typeof(string), // Title
                    typeof(string), // Genre
                    typeof(string), // Date
                    typeof(string), // Track
                    typeof(string), // Composer
                    typeof(string), // Performer
                    typeof(string), // Comment
                    typeof(int),    // Time
                    typeof(int),    // ID
                    typeof(bool));  // IsPlaying

            filter = new TreeModelFilter(tree, null);
            flt = new Filter(ref filter);
            if (MpcSharp.Settings.PlaylistShowFilter)
                PackStart(flt, false, false, 0);
            view.Model = filter;
            view.Selection.Mode = SelectionMode.Multiple;
            view.RulesHint = MpcSharp.Settings.UseRulesHint;
            TargetEntry[] targets = new TargetEntry[1];
            targets[0] = new TargetEntry("STRING", TargetFlags.Widget, 0);
            view.EnableModelDragSource(Gdk.ModifierType.Button1Mask, targets, Gdk.DragAction.Move);
            view.EnableModelDragDest(targets, Gdk.DragAction.Move);

            cell.Background = MpcSharp.Settings.PlaylistActiveColor;

            SetColumns();

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;

            // Watch gconf settings
            MpcSharp.Settings.PlaylistActiveColorChanged +=
                NewColorNotifyEventHandler;
            MpcSharp.Settings.PlaylistFormatChanged +=
                PlaylistFormatNotifyEventHandler;
            MpcSharp.Settings.CustomColumnFormatChanged +=
                CustomColumnFormatNotifyEventHandler;
            MpcSharp.Settings.UseRulesHintChanged +=
                UseRulesHintNotifyEventHandler;
            MpcSharp.Settings.PlaylistShowFilterChanged +=
                PlaylistShowFilterNotifyEventHandler;

        }

        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        public TabType Type {
            get {
                return TabType.Playlist;
            }
        }

        public string Title {
            get {
                return "Playlist";
            }
        }

        void NewColorNotifyEventHandler(object o, NotifyEventArgs args)
        {
            cell.Background = (string) args.Value;
            view.QueueDraw();
        }

        void PlaylistFormatNotifyEventHandler(object o, NotifyEventArgs args)
        {
            SetColumns();
        }

        void CustomColumnFormatNotifyEventHandler(object o, NotifyEventArgs args)
        {
            view.QueueDraw();
            // TODO: If cell height changes the view becomes corrupt. A toggle
            // of the column fixes it but there must be another way.
            //SetColumns();
        }
        void UseRulesHintNotifyEventHandler(object o, NotifyEventArgs args)
        {
            view.RulesHint = (bool) args.Value;
        }

        void PlaylistShowFilterNotifyEventHandler(object o, NotifyEventArgs args)
        {
            if ((bool) args.Value)
                PackStart(flt, false, false, 0);
            else
                Remove(flt);
            flt.ShowAll();

        }



        /*
         * The update delegate method for the playlist.
         * These things must be checked:
         *  - changes to the playlist
         *  - if a new song should be marked as 'playing'
         *  - maybe center the playlist
         */
        void Update()
        {
            int oldVersion = Connection.Instance.prevStatusInfo.PlaylistVersion;
            try
            {
                Connection.Instance.client.PlaylistChanges(oldVersion, PlChangesCallback);
                while (Connection.Instance.client.ReadResponse())
                {}

                // Some things must be updated if we have switched songs
                if (Connection.Instance.prevStatusInfo.Music.SongID != 
                        Connection.Instance.statusInfo.Music.SongID ||
                        Connection.Instance.prevStatusInfo.PlaylistVersion !=
                        Connection.Instance.statusInfo.PlaylistVersion)
                {
                    int songPos     = Connection.Instance.statusInfo.Music.Song;
                    int prevSongPos = Connection.Instance.prevStatusInfo.Music.Song;

                    // Unmark old
                    TreeIter iter;
                    if (tree.IterNChildren() > 0 && prevSongPos >= 0)
                    {
                        tree.GetIterFromString(out iter, Convert.ToString(prevSongPos));
                        if (tree.IterIsValid(iter))
                                tree.SetValue(iter, (int) StoreCols.IsPlaying, false); 
                    }
                    // Mark new
                    if (tree.IterNChildren() > 0 && songPos >= 0)
                    {
                        tree.GetIterFromString(out iter, Convert.ToString(songPos));
                        tree.SetValue(iter, (int) StoreCols.IsPlaying, true); 
                    }

                    // Possibly auto center playing
                    if (MpcSharp.Settings.AutoCenter)
                        Center();
                }
            }
            catch (MPD.MPDException e)
            {
                Connection.Instance.client.CleanCache();
                Console.WriteLine(e);
            }
        }

        /*
         * Callback executed when mpd is queried for changes to the playlist
         * since the last known playlist version. Incorporate the changes to
         * make the playlist stay in sync.
         */
        void PlChangesCallback(object result)
        {
            TreeIter iter;
            int curr_len = tree.IterNChildren();
            foreach (MusicInfo info in ((MusicCollection) result).Songs)
            {
                // Modify existing songs
                if (info.Pos < curr_len)
                {
                    tree.GetIterFromString(out iter, Convert.ToString(info.Pos));
                    // TODO: Figure out how ListStore.SetValist works instead
                    tree.SetValue(iter, (int) StoreCols.Filename, info.Filename);
                    tree.SetValue(iter, (int) StoreCols.Artist, info.Artist);
                    tree.SetValue(iter, (int) StoreCols.Album, info.Album);
                    tree.SetValue(iter, (int) StoreCols.Title, info.Title);
                    tree.SetValue(iter, (int) StoreCols.Genre, info.Genre);
                    tree.SetValue(iter, (int) StoreCols.Date, info.Date);
                    tree.SetValue(iter, (int) StoreCols.Track, info.Track);
                    tree.SetValue(iter, (int) StoreCols.Composer, info.Composer);
                    tree.SetValue(iter, (int) StoreCols.Performer, info.Performer);
                    tree.SetValue(iter, (int) StoreCols.Comment, info.Comment);
                    tree.SetValue(iter, (int) StoreCols.Time, info.Time);
                    tree.SetValue(iter, (int) StoreCols.ID, info.ID);
                    tree.SetValue(iter, (int) StoreCols.IsPlaying, false);
                }
                // Add new songs
                else

                    tree.AppendValues(
                            info.Filename,
                            info.Artist,
                            info.Album,
                            info.Title,
                            info.Genre,
                            info.Date,
                            info.Track,
                            info.Composer,
                            info.Performer,
                            info.Comment,
                            info.Time,
                            info.ID,
                            false);
            }

            // Remove songs
            int newLength = Connection.Instance.statusInfo.PlaylistLength;
            int currLength = tree.IterNChildren();
            while (currLength > newLength)
            {
                tree.GetIterFromString(out iter, Convert.ToString(currLength - 1));
                tree.Remove(ref iter);
                currLength = tree.IterNChildren();
            }
        }


        // Parse configured column string and add the columns to the TreeView
        void SetColumns()
        {
            // Remove all old columns
            foreach(TreeViewColumn col in view.Columns)
            {
                view.RemoveColumn(col);
            }

            // Append the new ones
            foreach (string pair in MpcSharp.Settings.PlaylistFormat)
            {
                if (pair == "")
                    continue;
                string name = pair.Split(':')[0];
                int width = Convert.ToInt32(pair.Split(':')[1]);

                TreeViewColumn tvc;

                if (name == "Custom")
                {
                    tvc = new TreeViewColumn(name, cell,
                            "background_set", StoreCols.IsPlaying);
                    tvc.SetCellDataFunc(cell,
                            new TreeCellDataFunc(CustomColumnCellDataFunc));
                }
                else if (name == "Time")
                {
                    tvc = new TreeViewColumn(name, cell,
                            "background_set", StoreCols.IsPlaying);
                    tvc.SetCellDataFunc(cell,
                            new TreeCellDataFunc(TimeColumnCellDataFunc));
                }
                else
                {
                    int col = (int) Enum.Parse(typeof(StoreCols), name);
                    tvc = new TreeViewColumn(name, cell,
                            "text", col, "background_set", StoreCols.IsPlaying);
                }
                tvc.FixedWidth = width;
                tvc.Sizing = TreeViewColumnSizing.Fixed;
                tvc.Resizable = true;
                view.AppendColumn(tvc);
            }

            // Show headers if we have more than one column
            view.HeadersVisible = view.Columns.Length > 1;
        }

        void TimeColumnCellDataFunc( TreeViewColumn tvc,
                CellRenderer cell, TreeModel model, TreeIter iter)
        {
            int time = (int) model.GetValue(iter, (int) StoreCols.Time);
            (cell as CellRendererText).Text = Format.TimeInSeconds(time);
        }

        void CustomColumnCellDataFunc( TreeViewColumn tvc,
                CellRenderer cell, TreeModel model, TreeIter iter)
        {
            // TODO: Inefficient. There must be a better way to do this ;)
            MusicInfo musicInfo = new MusicInfo(new Hashtable());
            musicInfo.Filename = (string) model.GetValue(iter, 0);
            musicInfo.Artist = (string) model.GetValue(iter, 1);
            musicInfo.Album = (string) model.GetValue(iter, 2);
            musicInfo.Title = (string) model.GetValue(iter, 3);
            musicInfo.Genre = (string) model.GetValue(iter, 4);
            musicInfo.Date = (string) model.GetValue(iter, 5);
            musicInfo.Track = (string) model.GetValue(iter, 6);
            musicInfo.Composer = (string) model.GetValue(iter, 7);
            musicInfo.Performer = (string) model.GetValue(iter, 8);
            musicInfo.Comment = (string) model.GetValue(iter, 9);
            musicInfo.Time = Convert.ToInt32(model.GetValue(iter, 10));
            Format formatter = new Format(MpcSharp.Settings.CustomColumnFormat);
            string text = formatter.MusicInfo(musicInfo);
            text = text.Replace(System.Environment.NewLine, "");
            text = text.Replace("\\n", System.Environment.NewLine);
            text = text.Replace("\\t", "\t");
            (cell as CellRendererText).Markup = text;
        }

        void ShowMenu()
        {
            bool selected = view.Selection.CountSelectedRows() > 0;
            gxml_menu.GetWidget("delete").Sensitive = selected;
            gxml_menu.GetWidget("crop").Sensitive = selected;

            ((Menu) gxml_menu.GetWidget("pl_popup")).Popup();
        }
        override protected bool OnPopupMenu()
        {
            ShowMenu();
            return true;
        }

        void Center()
        {
            view.ScrollToCell(
                    new TreePath(Convert.ToString(Connection.Instance.musicInfo.Pos)),
                    null, true, 0.5f, 0.5f);
        }

        /*
         * Main gui callbacks
         */
        private void OnRowActivated(object o, RowActivatedArgs args)
        {
            TreeIter iter = new TreeIter();
            int n = Convert.ToInt32(args.Path.ToString());
            filter.IterNthChild(out iter, n);
            int id = (int) filter.GetValue(iter, (int) StoreCols.ID);
            try
            {
                Connection.Instance.client.Play(id);
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException)
            {}
        }

        [GLib.ConnectBefore]
        private void OnButtonPress(object o, ButtonPressEventArgs args)
        {
            if (args.Event.Button == 3)
                args.RetVal = (o as TreeView).Selection.CountSelectedRows() > 1;
        }

        private void OnButtonRelease(object o, ButtonReleaseEventArgs args)
        {
            if (args.Event.Button == 3)
                ShowMenu();
            args.RetVal = true;
        }

        private void OnPopupMenu(object o, PopupMenuArgs args)
        {
            ShowMenu();
        }

        private void OnKeyPress(object o, KeyPressEventArgs args)
        {
            Gdk.EventKey e = args.Event;

            if (args.Event.Key == Gdk.Key.c)
            {
                Center();
                args.RetVal = true;
            }
            else if (e.Key == Gdk.Key.Delete)
            {
                gxml_menu.GetWidget("delete").Activate();
            }
        }

        [GLib.ConnectBefore]
        private void OnDragDrop(object o, DragDropArgs args)
        {
            // Get all the selected paths
            TreeModel model;
            TreePath[] selPaths;
            selPaths = view.Selection.GetSelectedRows(out model);

            // Transform them to id's because that's what we send to mpd
            TreeIter iter;
            ArrayList selIds = new ArrayList();
            foreach (TreePath path in selPaths)
            {
                model.GetIter(out iter, path);
                selIds.Add(model.GetValue(iter, (int) StoreCols.ID));
            }

            // Get the drop position
            TreePath dropPath;
            TreeViewDropPosition dropPos;
            if (!view.GetDestRowAtPos(args.X, args.Y, out dropPath, out dropPos))
                return;
            int newPos = dropPath.Indices[0];

            // Depending on the exact drop position, we may need to adjust the
            // new position by one
            if (newPos > selPaths[0].Indices[0])
            {
                if (dropPos == TreeViewDropPosition.Before)
                    newPos--;
            }
            else
            {
                if (dropPos == TreeViewDropPosition.After)
                {
                    newPos++;
                }
            }

            // If we are moving stuff up in the playlist we must move all
            // selected songs in reverse order to preserve the order within the
            // selection
            if (newPos < selPaths[0].Indices[0])
                selIds.Reverse();


            // Move them
            foreach (int i in selIds)
            {
                Connection.Instance.client.MoveId(i, newPos);
            }
            while (Connection.Instance.client.ReadResponse()) {}

            // Select the new position
            view.SetCursor(new TreePath(Convert.ToString(newPos)), view.Columns[0], false);
            
            // Stop event propagation here because the TreeModelFilter doesn't
            // have a default drag-drop handler which results in an ugly
            // warning.
            args.RetVal = true;
        }

        /*
         * Menu callbacks
         */
        private void OnMenuAdd(object o, EventArgs args)
        {
            MessageDialog md = new MessageDialog(Toplevel as Window,
                    DialogFlags.DestroyWithParent,
                    MessageType.Question,
                    ButtonsType.OkCancel,
                    "HTTP stream to add:");
            Entry entry = new Entry();
            md.VBox.PackStart(entry);
            md.VBox.ShowAll();
            ResponseType result = (ResponseType) md.Run();
            if (result == ResponseType.Ok)
            {
                try
                {
                    Connection.Instance.client.AddStream(entry.Text, false);
                    Connection.Instance.client.ReadResponse();
                }
                catch (NullReferenceException)
                {}
                catch (Exception e)
                {
                    Console.WriteLine(e);
                }
            }
            md.Destroy();
        }
        private void OnMenuSave(object o, EventArgs args)
        {
            MessageDialog md = new MessageDialog(Window.ListToplevels()[0],
                    DialogFlags.DestroyWithParent,
                    MessageType.Question,
                    ButtonsType.OkCancel,
                    "Filename of new playlist:");
            Entry entry = new Entry();
            md.VBox.PackStart(entry);
            md.VBox.ShowAll();
            ResponseType result = (ResponseType) md.Run();
            if (result == ResponseType.Ok)
            {
                try
                {
                    Connection.Instance.client.SavePlaylist(entry.Text);
                    Connection.Instance.client.ReadResponse();
                }
                catch (MPD.MPDException e)
                {
                    Console.WriteLine(e);
                }
                catch (NullReferenceException)
                {}
            }
            md.Destroy();
        }

        private void OnMenuDelete(object o, EventArgs args)
        {
            TreeModel model;
            TreePath[] paths = view.Selection.GetSelectedRows(out model);
            Delete(paths);
            if (paths.Length > 0 && view.Columns.Length > 0)
                view.SetCursor(paths[0], view.Columns[0], false);
        }

        private void OnMenuCrop(object o, EventArgs args)
        {
            // Add all selected IDs to a list
            ArrayList selArray = new ArrayList();
            TreeModel model;
            TreePath[] selRows = view.Selection.GetSelectedRows(out model);

            TreeIter iter;
            foreach(TreePath path in selRows)
            {
                filter.GetIter(out iter, path);
                selArray.Add((int) filter.GetValue(iter, (int) StoreCols.ID));
            }


            // Step trought whole playlist, deleting everything not in the
            // selected list
            if (filter.GetIterFirst(out iter))
            {
                try
                {
                    do
                    {
                        int id =  (int) filter.GetValue(iter, (int) StoreCols.ID);
                        if (selArray.BinarySearch(id) < 0)
                            Connection.Instance.client.RemoveSong(id);
                    } while (filter.IterNext(ref iter));
                    while (Connection.Instance.client.ReadResponse()) {}
                }
                catch (MPD.MPDException e)
                {
                    Console.WriteLine(e);
                }
                catch (NullReferenceException)
                {}
            }
        }
        private void OnMenuClear(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Clear();
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
        private void OnMenuShuffle(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Shuffle();
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        private void Delete(TreePath[] paths)
        {
            TreeIter iter;

            try
            {
                foreach(TreePath path in paths)
                {
                    filter.GetIter(out iter, path);
                    int id =  (int) filter.GetValue(iter, (int) StoreCols.ID);
                    Connection.Instance.client.RemoveSong(id);
                }
                while (Connection.Instance.client.ReadResponse()) {}
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        /*
         * Filter the playlist, making it show only items where the specified
         * metadata field matches the entered string.
         * The matching is case insensitive.
         */
         class Filter : HBox
         {
             Entry entry;
             ComboBox combo;
             TreeModelFilter treeModelFilter;

             public Filter(ref TreeModelFilter treeModelFilter) : base()
             {
                 this.treeModelFilter = treeModelFilter;

                 entry = new Entry();
                 combo = ComboBox.NewText();
                 foreach (string item in SearchType.GetNames(typeof(SearchType)))
                 {
                     // Only show items from 'SearchType' that are available in
                     // 'StoreCols'. The item 'Any' alone does not fulfill this
                     // requirement.
                     if (item != "Any")
                         combo.AppendText(item);
                 }

                 treeModelFilter.SetVisibleFunc(VisibleFunc, combo.Raw, null);

                 // Event handlers
                 entry.Changed += new EventHandler(OnEntryChanged);
                 combo.Changed += new EventHandler(OnComboChanged);

                 // Pack widgets
                 Label label = new Label("F_ilter");
                 label.MnemonicWidget = entry;
                 PackStart(label, false, false, 0);
                 PackStart(entry);
                 PackStart(combo, false, false, 0);
                 combo.Active = 0;

             }

             void OnEntryChanged(object o, EventArgs args)
             {
                 treeModelFilter.Refilter();
             }

             void OnComboChanged(object o, EventArgs args)
             {
                 treeModelFilter.Refilter();
             }

             bool VisibleFunc(TreeModel model, TreeIter iter)
             {
                 string what = entry.Text;
                 if (what == "")
                     return true;
                 try
                 {
                     string where = ((string) model.GetValue(
                                 iter, combo.Active)).ToLower();
                     return (where.IndexOf(what.ToLower()) >= 0);
                 }
                 catch (Exception)
                 {
                     return false;
                 }
             }
         }
    }
}
