/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

namespace MpcSharp
{
    using System;
    using Gtk;
    using MPD;

    class Playlists : TreeView, ITab
    {
        ListStore tree;
        [Glade.Widget] Menu popup;
        [Glade.Widget] MenuItem addMenuItem;
        [Glade.Widget] MenuItem playMenuItem;
        [Glade.Widget] MenuItem deleteMenuItem;
        [Glade.Widget] MenuItem refreshMenuItem;

        public Playlists() : base()
        {
            tree = new ListStore(typeof(string));
            Model = tree;

            CellRendererPixbuf cellPixbuf = new CellRendererPixbuf();
            cellPixbuf.StockId = Gtk.Stock.JustifyFill;
            AppendColumn("Type", cellPixbuf);
            AppendColumn("Name", new CellRendererText(), "text", 0);
            Selection.Mode = SelectionMode.Multiple;
            HeadersVisible = false;

            Glade.XML gxml_menu = Glade.XML.FromAssembly("playlists.glade", "popup", null);
            gxml_menu.Autoconnect(this);

            // Register event handlers
            RowActivated += new RowActivatedHandler(OnRowActivated);
            ButtonPressEvent += new ButtonPressEventHandler(OnButtonPress);
            ButtonReleaseEvent += new ButtonReleaseEventHandler(OnButtonRelease);

            // Register update delegate
            Connection.Instance.UpdateDelegates += Update;
        }

        public void UnregisterUpdateDelegate()
        {
            Connection.Instance.UpdateDelegates -= Update;
        }

        public TabType Type {
            get {
                return TabType.Playlists;
            }
        }

        public string Title {
            get {
                return "Playlists";
            }
        }

        void Update()
        {
            if (Connection.Instance.statsInfo.DBUpdateDate >
                    Connection.Instance.prevStatsInfo.DBUpdateDate)
                Reload();
        }

        void Reload()
        {
            tree.Clear();

            // Reload playlists
            Connection.Instance.client.ListPlaylists(ListPlaylistsCallback);
            while (Connection.Instance.client.ReadResponse())
            {}
        }

        void ListPlaylistsCallback(object result)
        {
            foreach (string playlist in (string[]) result)
            {
                tree.AppendValues(playlist);
            }
        }

        void EnqueueSelection()
        {
            TreeIter iter;
            // Return if nothing is selected
            if (!tree.GetIterFirst(out iter))
                return;

            TreeModel model;
            TreePath[] paths = Selection.GetSelectedRows(out model);

            foreach (TreePath path in paths)
            {
                model.GetIter(out iter, path);
                try
                {
                    string val = (string) model.GetValue(iter, 0);
                    Connection.Instance.client.LoadPlaylist(val);
                    Connection.Instance.client.ReadResponse();
                }
                catch (MPD.MPDException e)
                {
                    Console.WriteLine(e);
                }
                catch (NullReferenceException)
                {}
            }
        }

        void ShowMenu()
        {
            bool selected = Selection.CountSelectedRows() > 0;
            playMenuItem.Sensitive = selected;
            addMenuItem.Sensitive = selected;
            deleteMenuItem.Sensitive = selected;
            popup.Popup();
        }

        /*
         * Gui callbacks
         */
        void OnRowActivated(object o, RowActivatedArgs args)
        {
            playMenuItem.Activate();
        }

        [GLib.ConnectBefore]
        void OnButtonPress(object o, ButtonPressEventArgs args)
        {
            if (args.Event.Button == 3)
            {
                // If multiple items selected, don't deselect them upon menu popup
                args.RetVal =  Selection.CountSelectedRows() > 1;
            }
        }

        void OnButtonRelease(object o, ButtonReleaseEventArgs args)
        {
            if (args.Event.Button == 3)
            {
                ShowMenu();
            }
            args.RetVal = true;
        }

        /*
         * Menu callbacks
         */
        void OnAddActivate(object o, EventArgs args)
        {
            EnqueueSelection();
            try
            {
                Connection.Instance.client.Play();
                Connection.Instance.client.ReadResponse();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }

        void OnPlayActivate(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Clear();
                addMenuItem.Activate();
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }


        void OnUpdateActivate(object o, EventArgs args)
        {
            Console.WriteLine("Update");
            Reload();
        }

        void OnDeleteActivate(object o, EventArgs args)
        {
            TreeIter iter;
            TreeModel model;
            TreePath[] paths = Selection.GetSelectedRows(out model);
            // Loop through all selected rows
            foreach (TreePath path in paths)
            {
                tree.GetIter(out iter, path);
                string playlist = (string) tree.GetValue(iter, 0);
                // If it's a playlist, ask if it's ok to delete it
                MessageDialog md = new MessageDialog(Window.ListToplevels()[0],
                        DialogFlags.DestroyWithParent,
                        MessageType.Question,
                        ButtonsType.YesNo,
                        "Are you sure you want to delete '" + playlist + "'?");
                ResponseType result = (ResponseType) md.Run();
                if (result == ResponseType.Yes)
                {
                    try
                    {
                        Connection.Instance.client.RemovePlaylist(playlist);
                        Connection.Instance.client.ReadResponse();
                    }
                    catch (MPD.MPDException e)
                    {
                        Console.WriteLine(e);
                    }
                    catch (NullReferenceException)
                    {}
                }
                md.Destroy();
            }
            Reload();

        }
    }
}
