/*  Copyright (C) 2005 Magnus Bjernstad <bjernstad@gmail.com>
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

namespace MpcSharp
{
    using System;
    using Gtk;
    using GConf;
    using GConf.PropertyEditors;
    using MPD;

    /*
     * The settings editor presents a graphical interface for modifying the
     * configuration values stored in gconf. Other parts of the program can be
     * notified when a certain value has been changed.
     */
    class SettingsEditor
    {
        [Glade.Widget] TextView status_format;
        [Glade.Widget] TextView custom_column_format;
        [Glade.Widget] ColorButton playlist_active_colorbutton;
        [Glade.Widget] TreeView playlist_format_view;
        [Glade.Widget] TreeView gui_row_1_view;
        [Glade.Widget] TreeView gui_row_2_view;
        [Glade.Widget] TreeView outputs_view;
        [Glade.Widget] SpinButton crossfade;
        ListStore outputsModel;

        public SettingsEditor()
        {
            Glade.XML gxml = Glade.XML.FromAssembly("settings.glade", null, null);
            gxml.Autoconnect(this);

            GConf.PropertyEditors.EditorShell shell =
                new GConf.PropertyEditors.EditorShell (gxml);
            shell.Add(SettingKeys.Host, "host");
            shell.Add(SettingKeys.Port, "port");
            shell.Add(SettingKeys.Password, "password");
            shell.Add(SettingKeys.IncrementalSearch, "incremental_search");
            shell.Add(SettingKeys.MusicDirectory, "music_directory");
            shell.Add(SettingKeys.UseRulesHint, "use_rules_hint");
            shell.Add(SettingKeys.PlaylistShowFilter, "show_filter");
            shell.Add(SettingKeys.AutoCenter, "auto_center");

            PropertyEditorTextView peStatusFormat = new PropertyEditorTextView(
                    SettingKeys.StatusFormat, status_format);
            shell.Add(peStatusFormat);

            PropertyEditorTextView peCustomColumnFormat = new PropertyEditorTextView(
                    SettingKeys.CustomColumnFormat, custom_column_format);
            shell.Add(peCustomColumnFormat);

            PropertyEditorColorButton pePlaylistActiveColor = new PropertyEditorColorButton(
                    SettingKeys.PlaylistActiveColor, playlist_active_colorbutton);
            shell.Add(pePlaylistActiveColor);


            /*
             * PlaylistFormat
             */
            ListStore playlistFormatModel = new ListStore(typeof(bool), typeof(string), typeof(int));

            foreach (string name in Enum.GetNames(typeof(PlaylistColumn)))
            {
                int width = 100; // Default width
                bool visible = false; // Default visibility
                foreach (string pair in Settings.PlaylistFormat)
                {
                    string nameVis = pair.Split(':')[0];
                    if (nameVis == name)
                    {
                        width = Convert.ToInt32(pair.Split(':')[1]);
                        visible = true;
                    }
                }
                playlistFormatModel.AppendValues(visible, name, width);
            }
            playlist_format_view.Model = playlistFormatModel;


            CellRendererToggle cellVisible = new CellRendererToggle();
            CellRendererText cellType = new CellRendererText();
            CellRendererText cellWidth = new CellRendererText();

            cellWidth.Editable = true;

            TreeViewColumn tvc = new TreeViewColumn("Visible", cellVisible, "active", 0);
            playlist_format_view.AppendColumn(tvc);
            tvc = new TreeViewColumn("Type", cellType, "text", 1);
            playlist_format_view.AppendColumn(tvc);
            tvc = new TreeViewColumn("Width", cellWidth, "text", 2);
            playlist_format_view.AppendColumn(tvc);

            PropertyEditorPlaylistFormat pePlaylistFormat = new PropertyEditorPlaylistFormat(
                    SettingKeys.PlaylistFormat, playlist_format_view);
            shell.Add(pePlaylistFormat);

            /*
             * GuiRow1
             */
            ListStore guiRow1Model = new ListStore(typeof(bool), typeof(string));
            foreach(string name in Enum.GetNames(typeof(GuiWidget)))
            {
                bool visible = Array.IndexOf(Settings.GuiRow1, name) >= 0;
                guiRow1Model.AppendValues(visible, name);
            }
            gui_row_1_view.Model = guiRow1Model;

            gui_row_1_view.AppendColumn("Visible", new CellRendererToggle(), "active", 0);
            gui_row_1_view.AppendColumn("Widget", new CellRendererText(), "text", 1);

            PropertyEditorGuiRowFormat peGuiRow1Format = new PropertyEditorGuiRowFormat(
                    SettingKeys.GuiRow1, gui_row_1_view);
            shell.Add(peGuiRow1Format);

            /*
             * GuiRow2
             */
            ListStore guiRow2Model = new ListStore(typeof(bool), typeof(string));
            foreach (string name in Enum.GetNames(typeof(GuiWidget)))
            {
                bool visible = Array.IndexOf(Settings.GuiRow2, name) >= 0;
                guiRow2Model.AppendValues(visible, name);
            }
            gui_row_2_view.Model = guiRow2Model;

            gui_row_2_view.AppendColumn("Visible", new CellRendererToggle(), "active", 0);
            gui_row_2_view.AppendColumn("Widget", new CellRendererText(), "text", 1);

            PropertyEditorGuiRowFormat peGuiRow2Format = new PropertyEditorGuiRowFormat(
                    SettingKeys.GuiRow2, gui_row_2_view);
            shell.Add(peGuiRow2Format);

            /*
             * Outputs
             */
            outputsModel = new
                ListStore(typeof(bool), typeof(string), typeof(int));
            Connection.Instance.client.ListOutputs(ListOutputsCallback);
            while (Connection.Instance.client.ReadResponse())
            {}
            outputs_view.Model = outputsModel;

            CellRendererToggle cellOutputToggle = new CellRendererToggle();
            tvc = new TreeViewColumn("Active", cellOutputToggle, "active", 0);
            outputs_view.AppendColumn(tvc);
            tvc = new TreeViewColumn("Name", new CellRendererText(), "text", 1);
            outputs_view.AppendColumn(tvc);

            cellOutputToggle.Toggled += new ToggledHandler(OnOutputToggled);

            /*
             * Crossfade
             */
            crossfade.Value = Connection.Instance.statusInfo.Crossfade;
        }

        void ListOutputsCallback(object result)
        {
            outputsModel.Clear();
            foreach (OutputInfo output in ((OutputCollection) result).Outputs)
            {
                outputsModel.AppendValues(output.Enabled, output.Name, output.Id);
            }

        }

        void OnOutputToggled(object o, ToggledArgs args)
        {
            // Get the new values
            TreeIter iter;
            outputsModel.GetIter(out iter, new TreePath(args.Path));
            bool newValue = !(bool) outputsModel.GetValue(iter, 0);
            int id = (int) outputsModel.GetValue(iter, 2);

            // Try to enable/disable the output
            try
            {
                if (newValue)
                    Connection.Instance.client.EnableOutput(id);
                else
                    Connection.Instance.client.DisableOutput(id);
            }
            catch (MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}

            // List the outputs again to see what actually happened
            Connection.Instance.client.ListOutputs(ListOutputsCallback);
            while (Connection.Instance.client.ReadResponse())
            {}
        }


        void OnCrossfadeValueChanged(object o, EventArgs args)
        {
            try
            {
                Connection.Instance.client.Crossfade((o as SpinButton).ValueAsInt);
            }
            catch (MPD.MPDException e)
            {
                Console.WriteLine(e);
            }
            catch (NullReferenceException)
            {}
        }
    }
}

namespace GConf.PropertyEditors
{
    using Gtk;
    using System;
    using System.Collections;
    using System.Text.RegularExpressions;
    using System.Text;

    class PropertyEditorTextView : PropertyEditor
    {
        public PropertyEditorTextView(string key, TextView tv) : base(key, tv)
        {
        }

        protected override void ValueChanged(object sender, NotifyEventArgs args)
        {
            object val = args.Value;
            TextView tv = (TextView) Control;
            // Make this check to prevent an infinit loop
            if (tv.Buffer.Text != (string) val)
                tv.Buffer.Text = (string) val;
        }

        void Changed(object o, EventArgs args)
        {
            TextBuffer tb = (TextBuffer) o;
            Set((string) tb.Text);
        }

        protected override void ConnectHandlers()
        {
            TextView tv = (TextView) Control;
            tv.Buffer.Changed += new EventHandler(Changed);
        }
    }

    /*
     * Allows configuration of a color selected with a Gtk.ColorButton. The
     * value is stored in GConf as a hexadecimal string (eg #1199ff).
     */
    public class PropertyEditorColorButton : PropertyEditor
    {
        public PropertyEditorColorButton(string key, ColorButton cb) : base(key, cb)
        {
        }

        protected override void ValueChanged(object sender, NotifyEventArgs args)
        {
            object val = args.Value;
            if (val == null)
                return;
            Gdk.Color c = new Gdk.Color();
            Gdk.Color.Parse((string) val, ref c);
            ColorButton cb = (ColorButton) Control;
            cb.Color = c;
        }

        byte ToByte(uint val)
        {
            return (byte) (val >> 8);
        }

        void OnColorSet(object o, EventArgs args)
        {
            ColorButton cb = (ColorButton) Control;
            Gdk.Color c = cb.Color;
            Set(ColorToHex(c));
        }

        protected override void ConnectHandlers()
        {
            ColorButton cb = (ColorButton) Control;
            cb.ColorSet += new EventHandler(OnColorSet);
        }

        string ColorToHex (Gdk.Color c)
        {
            System.Text.StringBuilder s = new System.Text.StringBuilder ();
            s.Append ('#');

            ushort[] values = { c.Red, c.Green, c.Blue };
            char[] hexChars = {'0', '1', '2', '3', '4', '5', '6', '7',
                               '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

            foreach (ushort val in values)
            {
                byte rounded = (byte) (val >> 8);
                s.Append (hexChars[(rounded & 0xf0) >> 4]);
                s.Append (hexChars[rounded & 0x0f]);
            }
            return s.ToString ();
        }
    }

    public class PropertyEditorPlaylistFormat : PropertyEditor
    {
        public PropertyEditorPlaylistFormat(string key, TreeView view) : base(key, view)
        {
        }

        protected override void ValueChanged(object sender, NotifyEventArgs args)
        {
        }

        protected override void ConnectHandlers()
        {
            TreeView view = (TreeView) Control;
            ListStore model = (ListStore) view.Model;

            model.RowDeleted += new RowDeletedHandler(OnRowDeleted);
            (view.Columns[0].CellRenderers[0] as CellRendererToggle).Toggled +=
                new ToggledHandler(OnToggled);
            (view.Columns[2].CellRenderers[0] as CellRendererText).Edited +=
                new EditedHandler(OnEdited);
        }

        void OnToggled(object o, ToggledArgs args)
        {
            ListStore model = (ListStore) (Control as TreeView).Model;
            TreeIter iter;
            model.GetIter(out iter, new TreePath(args.Path));
            model.SetValue(iter, 0, !(bool)model.GetValue(iter, 0));
            Commit(model);
        }

        void OnEdited(object o, EditedArgs args)
        {
            ListStore model = (ListStore) (Control as TreeView).Model;
            TreeIter iter;
            model.GetIter(out iter, new TreePath(args.Path));
            try
            {
                model.SetValue(iter, 2, Convert.ToInt32(args.NewText));
                Commit(model);
            }
            catch (FormatException)
            {
                // NaN. Do nothing.
            }
        }
        
        void OnRowDeleted(object o, RowDeletedArgs args)
        {
            ListStore model = (ListStore) (Control as TreeView).Model;
            Commit(model);
        }

        /*
         * Store the values in the ListStore in GConf
         */
        void Commit(ListStore model)
        {
            ArrayList list = new ArrayList();
            TreeIter iter;
            model.GetIterFirst(out iter);
            do
            {
                bool show = (bool) model.GetValue(iter, 0);
                if (show)
                {
                    string name = (string) model.GetValue(iter, 1);
                    int width = (int) model.GetValue(iter, 2);

                    list.Add(name + ":" + width);
                }
            } while (model.IterNext(ref iter));

            if (list.Count > 0)
                Set(list.ToArray());
            else
            {
                string[] empty = new string[1];
                empty[0] = "";
                Set(empty);
            }
        }
    }
    

    public class PropertyEditorGuiRowFormat : PropertyEditor
    {
        public PropertyEditorGuiRowFormat(string key, TreeView view) : base(key, view)
        {
        }

        protected override void ValueChanged(object sender, NotifyEventArgs args)
        {
        }

        protected override void ConnectHandlers()
        {
            TreeView view = (TreeView) Control;
            ListStore model = (ListStore) view.Model;

            model.RowDeleted += new RowDeletedHandler(OnRowDeleted);
            (view.Columns[0].CellRenderers[0] as CellRendererToggle).Toggled +=
                new ToggledHandler(OnToggled);
        }

        void OnToggled(object o, ToggledArgs args)
        {
            ListStore model = (ListStore) (Control as TreeView).Model;
            TreeIter iter;
            model.GetIter(out iter, new TreePath(args.Path));
            model.SetValue(iter, 0, !(bool)model.GetValue(iter, 0));
            Commit(model);
        }

        void OnRowDeleted(object o, RowDeletedArgs args)
        {
            ListStore model = (ListStore) (Control as TreeView).Model;
            Commit(model);
        }

        /*
         * Store the values in the ListStore in GConf
         */
        void Commit(ListStore model)
        {
            ArrayList list = new ArrayList();
            TreeIter iter;
            model.GetIterFirst(out iter);
            do
            {
                bool show = (bool) model.GetValue(iter, 0);
                if (show)
                {
                    list.Add((string) model.GetValue(iter, 1));
                }
            } while (model.IterNext(ref iter));

            if (list.Count > 0)
                Set(list.ToArray());
            else
            {
                string[] empty = new string[1];
                empty[0] = "";
                Set(empty);
            }
        }
    }
}
